import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.Split;

import std.stdio;
import std.string;
import std.utf;

import util.Image;

import SubPanel1 ;
import SubPanel2 ;
import PPMutil ;

import Ccanvas ;
import Mandelbrot ;
import Coloring ;

ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }


class MainPanel : WindowObject {
  SubPanel1  subpanel1 ;
  SubPanel2  subpanel2 ;
  Split      splitbase ;

  Mandelbrot mandelbrot ;
  Coloring   mandelcolor ;
  
  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = LoadIcon(null , cast(wchar*)(IDI_APPLICATION));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    //    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = cast(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;
      
    case WM_CREATE:
      mandelbrot  = new Mandelbrot() ;
      mandelcolor = new Coloring() ;
      mandelbrot.setCanvasSize(640,480) ;
      
      splitbase = new Split("SplitBase",1) ;
      splitbase.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      splitbase.setPosition(0,0,256,256) ;
      splitbase.setSashPos(140,0,140) ;
      addObject( splitbase, cast(HMENU)(0), true, true) ;
      
      subpanel1 = new SubPanel1("SubPanel1","SubPanel1") ;
      subpanel1.setMandelBase(mandelbrot, mandelcolor) ;
      subpanel1.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      subpanel1.setPosition(0,0,256,256) ;
      splitbase.addObject(subpanel1, cast(HMENU)(0)) ;
      
      subpanel2 = new SubPanel2("SubPanel2","SubPanel2") ;
      subpanel2.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      subpanel2.setPosition(0,0,256,256) ;
      splitbase.addObject(subpanel2, cast(HMENU)(0)) ;

      subpanel1.setSubPanel2(subpanel2) ;
      
      subpanel2.graphic.setSubPanel1(subpanel1) ;
      subpanel2.graphic.setImageSize(mandelbrot.getCanvasWidth(),mandelbrot.getCanvasHeight()) ;
      subpanel2.graphic.setMandelPos(mandelbrot.ox, mandelbrot.oy, mandelbrot.zoom) ;
      
      mandelbrot.calc() ;
      auto canvas = mandelbrot.getCanvas() ;
      mandelcolor.coloring(canvas) ;
      uint *dstimg  = subpanel2.graphic.getImageBuffer() ;
      uint[] srcimg = mandelcolor.getImage() ;
      dstimg[0 .. srcimg.length] = srcimg[0..$] ;
      
      break ;
      
    case WM_MOVE:
      break ;
      
    case WM_SIZE:
      break;
      
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      break ;

    case WM_COMMAND:
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int setCanvasSize(int width, int height){
    mandelbrot.setCanvasSize(width,height) ;
    subpanel2.graphic.setImageSize(width,height) ;
    return(0) ;
  }

  int saveImage(string fname){
    if( mandelcolor !is null ){
      try{
	uint[] img = mandelcolor.getImage() ;
	
	//auto ppmutil = new PPMutil() ;
	//ppmutil.savefile(fname, img, mandelcolor.width, mandelcolor.height) ;

	auto image = new Image() ;
	image.setWidth( mandelcolor.width ) ;
	image.setHeight(mandelcolor.height) ;
	image.open( fname, "w" ) ;
	uint[] imgtmp ;
	imgtmp.length=mandelcolor.width ;
	for( int y=0 ; y<mandelcolor.height ; y++ ){
	  imgtmp[] = img[(y*mandelcolor.width) .. ((y+1)*mandelcolor.width)] ;
	  image.writeline( imgtmp ) ;
	}
	image.close() ;
      }catch(Exception e){
	return(-2) ;
      }
    }
    return(-1) ;
  }

}
