import std.stdio;
import std.stream;
import std.string;
import std.math;

import gmp;
import gmpp;

import Ccanvas;


class Mandelbrot
{
  mpf  ox    ;
  mpf  oy    ;
  mpf  zoom  ;
  Ccanvas canvas  ;

  this(){
    canvas = new Ccanvas ;
    canvas.width =640 ;
    canvas.height=640 ;
    ox = new mpf("0.0") ;
    oy = new mpf("0.0") ;
    zoom = new mpf("1.0") ;
  }

  int setCanvasSize(int width, int height){
    if( width<1 || height<1 ) return(-1) ;
    canvas.width =width  ;
    canvas.height=height ;
    canvas.alloc() ;
    return(0) ;
  }

  int setPosition(double ox, double oy){
    this.ox = new mpf(ox) ;
    this.oy = new mpf(oy) ;
    return(0) ;
  }

  int setPosition(mpf ox, mpf oy){
    this.ox = ox ;
    this.oy = oy ;
    return(0) ;
  }

  int setZoom(double zoom){
    this.zoom = new mpf(zoom) ;
    return(0) ;
  }

  int setZoom(mpf zoom){
    this.zoom = zoom ;
    return(0) ;
  }

  int setMaxCount(int maxcount){
    if( canvas !is null ){
      canvas.maxcount=maxcount ;
    }
    return(0) ;
  }

  int getCanvasWidth(){
    return(canvas.width) ;
  }
  
  int getCanvasHeight(){
    return(canvas.height) ;
  }

  int calc(){
    canvas.alloc() ;

    real rox = std.conv.to!(real)(ox.toString()) ;
    real roy = std.conv.to!(real)(oy.toString()) ;
    real rzm = std.conv.to!(real)(zoom.toString()) ;
    
    for( int y=0 ; y<canvas.height ; y++ ){
      for( int x=0 ; x<canvas.width ; x++ ){
	creal C  = (4.0*(cast(real)(x)-cast(real)(canvas.width )/2.0)/(rzm*canvas.width)+rox) +
		   (4.0*(cast(real)(y)-cast(real)(canvas.height)/2.0)/(rzm*canvas.width)+roy)*1.0i ;
	creal Z0 = 0.0 + 0.0i ;
	creal Z1 = 0.0 + 0.0i ;

	int count=0 ;
	for( int n=0 ; n<=canvas.maxcount ; n++ ){
	  Z1 = Z0 * Z0 + C ;
	  if( std.math.abs(Z1)>1000000.0 ) break ;
	  Z0 = Z1 ;
	  count++ ;
	}
	if( count<canvas.maxcount ){
	  canvas.array[y*canvas.width+x].count = count ;
	}else{
	  canvas.array[y*canvas.width+x].count = -1 ;
	}
      }
    }
    return(0) ;
  }

  int calcMPF(){
    canvas.alloc() ;

    uint precision=mpf_precision ;
    mpf_t gx    ; mpf_init2(&gx,precision) ;
    mpf_t gy    ; mpf_init2(&gy,precision) ;
    mpf_t gw    ; mpf_init2(&gw,precision) ;
    mpf_t gh    ; mpf_init2(&gh,precision) ;
    mpf_t  c_re ; mpf_init2( &c_re,precision) ;
    mpf_t  c_im ; mpf_init2( &c_im,precision) ;
    mpf_t z0_re ; mpf_init2(&z0_re,precision) ;
    mpf_t z0_im ; mpf_init2(&z0_im,precision) ;
    mpf_t z1_re ; mpf_init2(&z1_re,precision) ;
    mpf_t z1_im ; mpf_init2(&z1_im,precision) ;
    mpf_t th    ; mpf_init2(&th,precision) ;
    
    mpf_t tmp1  ; mpf_init2(&tmp1,precision) ;
    mpf_t tmp2  ; mpf_init2(&tmp2,precision) ;
    mpf_t tmp3  ; mpf_init2(&tmp3,precision) ;
    
    mpf_set_si( &gw, canvas.width  ) ;
    mpf_set_si( &gh, canvas.height ) ;
    //mpf_set_str(&ox, (("-0.67022214498468461443").dup).ptr, 10 ) ;
    //mpf_set_str(&oy, (("-0.45803615162270586636").dup).ptr, 10 ) ;
    //mpf_set_str(&gz, (("10000000000000000000.0").dup).ptr, 10 ) ; //mpf_set_str(&gz, (("25600000000000000000.0").dup).ptr, 10 ) ;
    
    for( int y=0 ; y<canvas.height ; y++ ){
      for( int x=0 ; x<canvas.width ; x++ ){
	mpf_set_si( &gx, x ) ; //scope gx = new mpf(x) ;
	mpf_set_si( &gy, y ) ; //scope gy = new mpf(y) ;
	
	//scope mpf c_re ; c_re=(4*(gx-gw/2))/(gz*gw)+ox ;
	mpf_div_ui( &tmp1, &gw, 2    ) ; // gw/2
	mpf_sub(    &tmp2, &gx, &tmp1) ; // (gx - gw/2)
	mpf_mul_ui( &tmp2, &tmp2, 4  ) ; // (4*(gx - gw/2))
	mpf_mul( &tmp3, &zoom.v, &gw ) ; // (gz*gw)
	mpf_div( &c_re, &tmp2, &tmp3 ) ; // (4*(gx-gw/2))/(gz*gw)
	mpf_add( &c_re, &c_re, &ox.v ) ; // (4*(gx-gw/2))/(gz*gw)+ox
	//scope mpf c_im ; c_im=(4*(gy-gh/2))/(gz*gw)+oy ;
	mpf_div_ui( &tmp1, &gh, 2    ) ; // gh/2
	mpf_sub(    &tmp2, &gy, &tmp1) ; // (gy - gh/2)
	mpf_mul_ui( &tmp2, &tmp2, 4  ) ; // (4*(gy - gh/2))
	mpf_div( &c_im, &tmp2, &tmp3 ) ; // (4*(gy-gh/2))/(gz*gw)
	mpf_add( &c_im, &c_im, &oy.v ) ; // (4*(gy-gh/2))/(gz*gw)+oy
	//scope  C  = new mpfComplex(c_re,c_im) ;
	
	//scope  Z0 = new mpfComplex(0, 0) ;
	mpf_set_si( &z0_re, 0 ) ;
	mpf_set_si( &z0_im, 0 ) ;
	
	mpf_set_d( &th, 1000000.0 ) ; //scope  TH = new mpf(1000000.0) ;
	
	int count=0 ;
	for( int n=0 ; n<canvas.maxcount ; n++ ){
	  //mpfComplex Z1 = Z0 * Z0 + C ;
	  mpf_mul( &z1_re , &z0_re, &z0_re ) ;
	  mpf_mul( &tmp1  , &z0_im, &z0_im ) ;
	  mpf_sub( &z1_re , &z1_re, &tmp1  ) ;
	  mpf_add( &z1_re , &z1_re, &c_re  ) ;
	  mpf_mul( &z1_im , &z0_re, &z0_im ) ;
	  mpf_mul( &tmp1  , &z0_re, &z0_im ) ;
	  mpf_add( &z1_im , &z1_im, &tmp1  ) ;
	  mpf_add( &z1_im , &z1_im, &c_im  ) ;
	  
	  mpf_mul( &tmp1, &z1_re, &z1_re ) ;
	  mpf_mul( &tmp2, &z1_im, &z1_im ) ;
	  mpf_add( &tmp1, &tmp1, &tmp2 ) ;
	  mpf_sqrt( &tmp1, &tmp1 ) ;
	  int z1abs_gt_th = mpf_cmp( &tmp1, &th ) ;
	  if( z1abs_gt_th>0 ) break ;
	  mpf_set(&z0_re, &z1_re) ;
	  mpf_set(&z0_im, &z1_im) ;
	  count++ ;
	}
	if( count<canvas.maxcount ){
	  canvas.array[y*canvas.width+x].count = count ;
	}else{
	  canvas.array[y*canvas.width+x].count = -1 ;
	}
      }
    }
    mpf_clear(&gx) ;
    mpf_clear(&gy) ;
    mpf_clear(&gw) ;
    mpf_clear(&gh) ;
    mpf_clear( &c_re) ;
    mpf_clear( &c_im) ;
    mpf_clear(&z0_re) ;
    mpf_clear(&z0_im) ;
    mpf_clear(&z1_re) ;
    mpf_clear(&z1_im) ;
    mpf_clear(&th) ;
    mpf_clear(&tmp1) ;
    mpf_clear(&tmp2) ;
    mpf_clear(&tmp3) ;

    return(0) ;
  }

  Ccanvas getCanvas(){
    return(canvas) ;
  }

}
