import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.Scrollbar;
import win32frame.TextBox;
import win32frame.UpDownCtrl;
import win32frame.StaticControl;


import std.stdio;
import std.string;
import std.utf;
import std.conv;
import std.math;

import gmpp;

ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }

enum CMD{
  UPDN  =0x301,
  TXTBOX=0x302
} ;
enum SPINMODE{
  LINEAR=0,
  BY    =1
} ;

class Spin : WindowObject {
  HGLRC     hRC;

  UpDownCtrl udcntl ;
  bool       update_text ;
  
  TextBox   textbox ;
  string    textfmt ;
  mpf       cur, max, min, step ;
  SPINMODE  mode ;
  int       scrW = 12 ;
  bool      dispmpf ;

  StaticControl label ;
  string    labelStr = "" ;
  int       labelW   = 0 ;
  
  this(string name, string ClassName, bool dispmpf=false) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = cast(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;

    cur = new mpf("0.0") ;
    min = new mpf("1.0") ;
    max = new mpf("-1.0") ;
    step= new mpf("1.0") ;
    mode=SPINMODE.LINEAR ;
    this.dispmpf=dispmpf ;
    if( dispmpf==false ){
      textfmt="%f" ;
    }else{
      textfmt="%s" ;
    }
    update_text=false ;
  }

  int setPosition(int x, int y, int w, int h){
    winX = x ;
    winY = y ;
    winW = w ;
    winH = h ;
    return(0) ;
  }

  int setInfo(double cur, double min, double max, double step, SPINMODE md=SPINMODE.LINEAR){
    this.cur = new mpf(cur) ;
    this.min = new mpf(min) ;
    this.max = new mpf(max) ;
    this.step= new mpf(step) ;
    this.mode= md  ;
    return(0) ;
  }

  int setInfo(string cur, string min, string max, string step, SPINMODE md=SPINMODE.LINEAR){
    this.cur = new mpf(cur) ;
    this.min = new mpf(min) ;
    this.max = new mpf(max) ;
    this.step= new mpf(step) ;
    this.mode= md  ;
    return(0) ;
  }

  int setValue(double dcur){
    auto cur = new mpf(dcur) ;
    if( min<=max ){
      cur = (cur<min) ? min : (cur>max) ? max : cur ;
    }
    if( textbox.getHWND() != null ){
      if( dispmpf==false ){
	textbox.setWindowText(format(textfmt,std.conv.to!(double)(cur.toString()))) ;
      }else{
	textbox.setWindowText(format(textfmt,cur)) ;
      }
    }
    this.cur = cur ;
    return(0) ;
  }
  
  int setValue(mpf cur){
    if( min<=max ){
      cur = (cur<min) ? min : (cur>max) ? max : cur ;
    }
    if( textbox.getHWND() != null ){
      if( dispmpf==false ){
	textbox.setWindowText(format(textfmt,std.conv.to!(double)(cur.toString()))) ;
      }else{
	textbox.setWindowText(format(textfmt,cur)) ;
      }
    }
    this.cur = cur ;
    return(0) ;
  }

  real getValue(){
    return( std.conv.to!(real)(cur.toString()) ) ;
  }

  mpf getMpfValue(){
    return( cur ) ;
  }

  int setTextFormat(string fmt){
    textfmt=fmt ;
    return(0) ;
  }

  int setTextFont(HFONT hfont){
    if( getHWND() != null ){
      return(textbox.setFont(hfont)) ;
    }
    return(-1) ;
  }

  int setLabel(string label, int width){
    this.labelStr  = label ;
    if( width>=0 ){
      this.labelW = width ;
    }
    return(0) ;
  }

  int setLabelFont(HFONT hfont){
    if( getHWND() != null ){
      return(label.setFont(hfont)) ;
    }
    return(-1) ;
  }
  
  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;
      
    case WM_CREATE:
      udcntl = new UpDownCtrl("updowncontrol") ;
      udcntl.setPosition(winW-scrW,winY,scrW,winH) ;
      addObject(udcntl, cast(HMENU)(CMD.UPDN), false, false) ;
      udcntl.setRange(0,100) ;
      udcntl.setPos(50) ;
      string defaultText ;
      if( dispmpf==false ){
	defaultText = format(textfmt,std.conv.to!(double)(cur.toString())) ;
      }else{
	defaultText = format(textfmt,cur) ;
      }
      textbox= new TextBox(defaultText) ;
      textbox.addExStyle(WS_EX_CLIENTEDGE|WS_EX_STATICEDGE ) ;
      textbox.addStyle(WS_BORDER | ES_RIGHT | ES_AUTOHSCROLL) ;
      textbox.setPosition(winX+labelW,winY,(winW-scrW-labelW),winH);
      addObject(textbox, cast(HMENU)(CMD.TXTBOX), false, false) ;
      label = new StaticControl(labelStr) ;
      label.addStyle(SS_RIGHT) ;
      label.setPosition(winX,winY,labelW,winH) ;
      addObject(label, null) ;
      break ;

    case WM_SIZE :
      if( (udcntl !is null) && (textbox !is null) && (label !is null) ){
	MoveWindow(  udcntl.getHWND(), winW-scrW, 0,             scrW,winH,true) ;
       	MoveWindow( textbox.getHWND(), labelW,    0, winW-scrW-labelW,winH,true) ;
       	MoveWindow(   label.getHWND(),      0,    0,           labelW,winH,true) ;
      }
      break ;

    case WM_VSCROLL:
      bool changed=false ;
      auto d = new mpf("0.0") ;
      if( HIWORD(wp)<50 ){
	if( mode == SPINMODE.LINEAR ){
	  d =  step * -(HIWORD(wp)-50);
	}else{
	  d =  step ;
	}
	changed=true ;
      }else if( HIWORD(wp)>50 ){
	if( mode == SPINMODE.LINEAR ){
	  d = -step *  (HIWORD(wp)-50);
	}else{
	  
	  d = 1/step ;
	}
	changed=true ;
      }
      udcntl.setPos(50) ;
      if( changed==true ){
	SetFocus(hwnd) ;
	if( mode == SPINMODE.LINEAR ){
	  cur = cur + d ;
	}else{
	  cur = cur * d ;
	}
	if( min<=max ){
	  cur = (cur<min) ? min : (cur>max) ? max : cur ;
	}
	if( dispmpf==false ){
	  textbox.setWindowText(format(textfmt,std.conv.to!(double)(cur.toString()))) ;
	}else{
	  textbox.setWindowText(format(textfmt,cur)) ;
	}
	sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), 0) ;
      }
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.TXTBOX:
	if( HIWORD(wp)==EN_KILLFOCUS ){
	  string val=textbox.getWindowText() ;
	  if( isNumeric(val)==true ){
	    cur = new mpf(val) ;
	    if( min<=max ){
	      cur = (cur<min) ? min : (cur>max) ? max : cur ;
	    }
	    if( dispmpf==false ){
	      textbox.setWindowText(format(textfmt,std.conv.to!(double)(cur.toString()))) ;
	    }else{
	      textbox.setWindowText(format(textfmt,cur)) ;
	    }
	    sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), 0) ;
	  }
	}
	break ;
      default :	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
