import std.stdio;
import std.string;
import std.regex;

struct JPG_HANDLE{
  uint *a ;
  uint *b ;
  uint *c ;
  uint *d ;
}

struct PNG_HANDLE{
  uint *a ;
  uint *b ;
  uint *c ;
  uint *d ;
  int   A ;
  int   B ;
}

enum FileOPstat{
  IDLE,
  JPGRDOPEN,
  PNGRDOPEN,
  JPGWTOPEN,
  PNGWTOPEN
}

  
extern (C) int readinfo_png( char *file, int *width, int *height, int *bpp ) ;
extern (C) int load_png_to32bitBuffer( char *file, int width, int height, int bpp, char* image ) ;
//extern (C) char* load_png(char *file, int *width, int *height, int *bpp);
extern (C) PNG_HANDLE *load_png_preproc( char *file ) ;
extern (C) int load_png_getline( uint *img, PNG_HANDLE *png_handle ) ;
extern (C) int load_png_postproc( PNG_HANDLE *png_handle ) ;

extern (C) PNG_HANDLE *save_png_preproc( char *file, int width, int height ) ;
extern (C) int save_png_putline( uint *img, PNG_HANDLE *png_handle ) ;
extern (C) int save_png_postproc( PNG_HANDLE *png_handle ) ;

extern (C) int readinfo_jpg( char *file, int *width, int *height, int *bpp );
extern (C) int load_jpg_to32bitBuffer( char *file, int width, int height, int bpp, char* image ) ;
//extern (C) char* load_jpg(char *file, int *width, int *height, int *bpp);
extern (C) JPG_HANDLE *load_jpg_preproc( char *file ) ;
extern (C) int load_jpg_getline( uint *img, JPG_HANDLE *jpg_handle ) ;
extern (C) int load_jpg_postproc( JPG_HANDLE *jpg_handle ) ;

class Image {
  int         width  ;
  int         height ;
  int         bpp    ;
  int         curline;
  JPG_HANDLE *jpg_handle ;
  PNG_HANDLE *png_handle ;
  FileOPstat  state  ;
  
  this(){
    width  = 0 ;
    height = 0 ;
    bpp    = 0 ;
    state  = FileOPstat.IDLE ;
    jpg_handle = null ;
  }

  ~this(){
    close() ;
  }

  int getWidth(){
    return(width) ;
  }
  
  int getHeight(){
    return(height) ;
  }
  
  int getBPP(){
    return(bpp) ;
  }
  
  int setWidth(int width){
    this.width=width ;
    return(0) ;
  }
  
  int setHeight(int height){
    this.height=height ;
    return(0) ;
  }
  
  int setBPP(int bpp){
    this.bpp=bpp ;
    return(0) ;
  }
  
  int open( string fname, string mode="r" ){
    int    w,h,b ;
    auto jpgext = regex(".jpg$","i") ;
    auto pngext = regex(".png$","i") ;

    if( state!=FileOPstat.IDLE ){
      close() ;
    }
      
    if( (match(fname,jpgext)).empty==false ){
      if( mode == "r" ){
	if( readinfo_jpg(cast(char*)(fname ~ "\0"), &w, &h, &b)!=0 ){
	  return(-1) ;
	}
	jpg_handle = load_jpg_preproc(cast(char*)(fname ~ "\0")) ;
	if( jpg_handle==null ) return(-1) ;
	state  = FileOPstat.JPGRDOPEN ;
	width  = w ;
	height = h ;
	bpp    = b ;
	curline= 0 ;
      }
      return(0) ;
    }

    if( match(fname,pngext).empty==false ){
      if( mode == "r" ){
	if( readinfo_png(cast(char*)(fname ~ "\0"), &w, &h, &b)!=0 ){
	  return(-1) ;
	}
	png_handle = load_png_preproc(cast(char*)(fname ~ "\0")) ;
	if( png_handle==null ) return(-1) ;
	state  = FileOPstat.PNGRDOPEN ;
	width  = w ;
	height = h ;
	bpp    = b ;
	curline= 0 ;
      }else{
	if( width<=0 || height<=0 ) return(-1) ;
	png_handle = save_png_preproc(cast(char*)(fname ~ "\0"), width, height) ;
	if( png_handle==null ) return(-1) ;
	state  = FileOPstat.PNGWTOPEN ;
	curline= 0 ;
      }
      return(0) ;
    }
    writef("unmatch ext(png or jpg) (%s)\n", fname) ;
    
    return(-1) ;
  }

  uint[] readline(int lnum=1, bool reverse_order=false){
    uint[] img ;

    switch( state ){
    case FileOPstat.JPGRDOPEN :
      img.length=width*lnum ;
      for( int i=0 ; i<lnum ; i++ ){
	load_jpg_getline( &img[i*width], jpg_handle) ;
	curline++ ;
      }
      break ;
    case FileOPstat.PNGRDOPEN :
      img.length=width*lnum ;
      for( int i=0 ; i<lnum ; i++ ){
	load_png_getline( &img[i*width], png_handle) ;
	curline++ ;
      }
      break ;
    default :
      break ;
    }

    if( reverse_order==true ){
      for( int i=0 ; i<img.length ; i++ ){
	img[i] = (img[i]&0xff00ff00) |  ((img[i]&0x00ff0000)>>16) |  ((img[i]&0x000000ff)<<16) ;
      }
    }
    
    return(img) ;
  }

  int writeline(uint[] img, int lnum=1, bool reverse_order=false){
    if( reverse_order==true ){
      for( int i=0 ; i<img.length ; i++ ){
	img[i] = (img[i]&0xff00ff00) |  ((img[i]&0x00ff0000)>>16) |  ((img[i]&0x000000ff)<<16) ;
      }
    }
    switch( state ){
    case FileOPstat.PNGWTOPEN :
      for( int i=0 ; i<lnum ; i++ ){
	save_png_putline( &img[i*width], png_handle) ;
	curline++ ;
      }
      break ;
    default :
      return(-1) ;
      break ;
    }
    
    return(0) ;
  }

  int close(){
    switch( state ){
    case FileOPstat.JPGRDOPEN :
      load_jpg_postproc( jpg_handle ) ;
      jpg_handle=null ;
      break ;
    case FileOPstat.PNGRDOPEN :
      load_png_postproc( png_handle ) ;
      png_handle=null ;
      break ;
    case FileOPstat.PNGWTOPEN :
      save_png_postproc( png_handle ) ;
      png_handle=null ;
      break ;
    default :
      break ;
    }
    state     =FileOPstat.IDLE ;
    width     = 0 ;
    height    = 0 ;
    bpp       = 0 ;
    curline   = 0 ;
    
    return(0) ;
  }
  
}

version (CLASS_DEBUG){
import std.stream;
  
int main(string[] args)
{
  Image img = new Image() ;

  if( args.length!=2 ){
    writef("Usage : %s filename\n",args[0]) ;
    return(0) ;
  }
  
  if( img.open(args[1])==0 ){
    FileMode     fmd ;
    BufferedFile outfile ;
    uint[]       iimg ;
    char[]       cimg ;

    outfile = new BufferedFile() ;
    outfile.create("test.ppm") ;
    outfile.writef("P6\n%d %d\n255\n",img.getWidth(),img.getHeight()) ;
    cimg.length=img.getWidth()*3 ;
    for( int i=0 ; i<img.getHeight() ; i++ ){
      iimg = img.readline() ;
      for( int j=0 ; j<iimg.length ; j++ ){
	cimg[j*3+0] = (iimg[j] & 0x00ff0000)>>16 ;
	cimg[j*3+1] = (iimg[j] & 0x0000ff00)>>8  ;
	cimg[j*3+2] = (iimg[j] & 0x000000ff)     ;
      }
      outfile.writeString(cimg) ;
    }
    outfile.close() ;
    
  }else{
    writef("file open error (%s)\n",args[1]) ;
  }
  return(0) ;
}
}
  
